import asyncio
import json
from enum import StrEnum

from curl_cffi import requests

from .session import ChaseSession
from .urls import execute_order, get_headers, order_info, order_page, order_status, validate_order


class PriceType(StrEnum):
    """Contains the valid price types for an order."""

    LIMIT = "LIMIT"
    MARKET = "MARKET"
    STOP = "STOP"
    STOP_LIMIT = "STOP_LIMIT"


class Duration(StrEnum):
    """Contains the valid durations for an order."""

    DAY = "DAY"
    GTC = "GOOD_TILL_CANCELLED"
    ON_OPEN = "ON_THE_OPEN"
    ON_CLOSE = "ON_THE_CLOSE"
    I_C = "IMMEDIATE_OR_CANCEL"


class OrderSide(StrEnum):
    """Contains the valid order types for an order."""

    BUY = "BUY"
    SELL = "SELL"
    SELL_ALL = "SELL_ALL"


class TypeCode(StrEnum):
    """Contains the valid order types for an order."""

    CASH = "CASH"
    MARGIN = "MARGIN"


class Order:
    """Contains information about an order and provides methods to place orders.

    Also contains a method to place an order.
    """

    def __init__(self, session: ChaseSession, accept_warning: bool = True) -> None:  # noqa: FBT001
        """Initialize an Order instance.

        Args:
            session (ChaseSession): The Chase session to use for order operations.
            accept_warning (bool, optional): Whether to accept warnings during order placement.
                                           Defaults to True.

        """
        self.session = session
        self.accept_warning = accept_warning
        self.order_number: str = ""

    def place_order(  # noqa: PLR0913, PLR0917
        self,
        account_id: str,
        quantity: int,
        price_type: PriceType,
        symbol: str,
        duration: Duration,
        order_type: OrderSide,
        limit_price: float = 0.00,
        stop_price: float = 0.00,
        after_hours: bool = True,  # noqa: FBT001, FBT002
        dry_run: bool = True,  # noqa: FBT001, FBT002
    ) -> dict:
        """Build and place an order.

        :attr: 'order_confirmation`
        contains the order confirmation data after order placement.

        Args:
            account_id (str): Account number of the account to place the order in.
            quantity (int): The number of shares to buy.
            price_type (PriceType): Price Type i.e. LIMIT, MARKET, STOP, etc.
            symbol (str): Ticker to place the order for.
            duration (Duration): Duration of the order i.e. DAY, GT90, etc.
            order_type (OrderSide): Type of order i.e. BUY, SELL, SELL_ALL.
            limit_price (float, optional): The price to buy the shares at. Defaults to 0.00.
            stop_price (float, optional): The price to buy the shares at. Defaults to 0.00.
            after_hours (bool, optional): Whether you want to place the order after hours. Defaults to True.
            dry_run (bool, optional): Whether you want the order to be placed or not.
                                      Defaults to True.

        Returns:
            Order:order_confirmation: Dictionary containing the order confirmation data.

        """
        return self.session.loop.run_until_complete(
            self._place_order_async(
                account_id,
                quantity,
                price_type,
                symbol,
                duration,
                order_type,
                limit_price=limit_price,
                stop_price=stop_price,
                after_hours=after_hours,
                dry_run=dry_run,
            ),
        )

    async def _place_order_async(  # noqa: PLR0913, PLR0917
        self,
        account_id: str,
        quantity: int,
        price_type: PriceType,
        symbol: str,
        duration: Duration,
        order_type: OrderSide,
        limit_price: float = 0.00,
        stop_price: float = 0.00,
        after_hours: bool = True,  # noqa: FBT001, FBT002
        dry_run: bool = True, # noqa: FBT001, FBT002
    ) -> dict:
        """Async implementation of place_order.

        Returns:
            dict: Dictionary containing order validation/confirmation data with keys
                  "ORDER INVALID", "ORDER VALIDATION", and "ORDER CONFIRMATION".

        """
        await self.session.page.get(order_page())

        order_messages = {
            "ORDER INVALID": "",
            "ORDER VALIDATION": "",
            "ORDER CONFIRMATION": "",
        }

        # Implement API calls instead of browser interactions where possible.
        # Thanks @OSSY!
        headers = get_headers()
        cookies = await self.session.browser.cookies.get_all()
        cookies_dict = {c.name: c.value for c in cookies}

        order_payload = {
           "accountIdentifier": int(account_id),
            # do i need this for market?
            "marketPriceAmount": limit_price,
            "orderQuantity": quantity,
            "accountTypeCode": "CASH",
            "timeInForceCode": duration,
            "securitySymbolCode": symbol,
            "tradeChannelName": "DESKTOP",
            "dollarBasedTradingEligibleIndicator": False,
            "orderTypeCode": price_type,
        }

        url_validate = validate_order(order_type=order_type)
        url_execute = execute_order(order_type=order_type)

        if order_type == "BUY":
            order_payload["tradeActionName"] = "BUY"
        elif order_type == "SELL":
            order_payload["tradeActionName"] = "SELL"
        elif order_type == "SELL_ALL":
            order_payload["tradeActionName"] = "SELL_ALL"

        if price_type == "LIMIT":
            order_payload["limitPriceAmount"] = limit_price
        elif price_type == "MARKET":
            if duration not in {"DAY", "ON_THE_CLOSE"}:
                order_messages["ORDER INVALID"] = (
                    "Market orders must be DAY or ON THE CLOSE."
                )
                return order_messages
        elif price_type == "MARKET ON CLOSE":
            pass
        elif price_type in {"STOP", "STOP_LIMIT"} and duration not in {"DAY", "GOOD_TILL_CANCELLED"}:
            order_messages["ORDER INVALID"] = (
                "Stop orders must be DAY or GOOD TILL CANCELLED."
            )
            return order_messages

        if price_type in {"LIMIT", "STOP_LIMIT"}:
            order_payload["limitPriceAmount"] = limit_price
        if price_type in {"STOP", "STOP_LIMIT"}:
            order_payload["stopPriceAmount"] = stop_price

        if duration == "DAY":
            order_payload["timeInForceCode"] = "DAY"
        elif duration == "GOOD_TILL_CANCELLED":
            order_payload["timeInForceCode"] = "GOOD_TILL_CANCELLED"
        elif duration == "ON_THE_OPEN":
            order_payload["timeInForceCode"] = "ON_THE_OPEN"
        elif duration == "ON_THE_CLOSE":
            order_payload["timeInForceCode"] = "ON_THE_CLOSE"
        elif duration == "IMMEDIATE_OR_CANCEL":
            order_payload["timeInForceCode"] = "IMMEDIATE_OR_CANCEL"

        try:
            # STEP 1: VALIDATION

            resp_val = requests.post(url_validate, headers=headers, cookies=cookies_dict, json=order_payload, impersonate="chrome")

            if resp_val.status_code != 200:
                order_messages["ORDER INVALID"] = f"Validation Failed ({resp_val.status_code}): {resp_val.text}"

            val_data = resp_val.json()

            error_msgs = val_data.get("tradeErrorMessages", [])
            order_messages["ORDER INVALID"] = error_msgs

            if order_messages["ORDER INVALID"]:
                return order_messages

            exchange_id = val_data.get("financialInformationExchangeSystemOrderIdentifier")

            if dry_run:
                order_messages["ORDER VALIDATION"] = val_data
                return order_messages
            if not exchange_id:
                order_messages["ORDER VALIDATION"] = val_data
        except Exception as e:
            order_messages["ORDER INVALID"] = f"Execution Exception: {e}"

        try:
            # STEP 2: EXECUTION
            payload_execute = order_payload.copy()
            payload_execute["financialInformationExchangeSystemOrderIdentifier"] = exchange_id

            resp_exec = requests.post(url_execute, headers=headers, cookies=cookies_dict, json=payload_execute, impersonate="chrome")

            if resp_exec.status_code != 200:
                order_messages["ORDER INVALID"] = f"Execution Failed ({resp_exec.status_code}): {resp_exec.text}"
            elif resp_exec.status_code == 200:
                order_messages["ORDER VALIDATION"] = val_data
            exec_data = resp_exec.json()
            # order_id = exec_data.get("orderIdentifier")
            order_messages["ORDER CONFIRMATION"] = exec_data
        except Exception as e:
            order_messages["ORDER INVALID"] = f"Execution Exception: {e}"
        return order_messages

    def get_order_statuses(self, account_id: str) -> str:
        """Retrieve the statuses of all recent orders placed.

        This method navigates to the order status page and scrapes the status of each order.
        It returns a list of dictionaries, where each dictionary represents an order and contains
        the order number and status.

        Returns:
            list[dict] | None: A list of dictionaries, where each dictionary contains 'order_number'
            and 'status' keys on success, or None if an error occurred while retrieving statuses.

        """
        return self.session.loop.run_until_complete(
            self._get_order_statuses_async(account_id),
        )

    async def _get_order_statuses_async(self, account_id: str) -> str:
        """Async implementation of get_order_statuses.

        Returns:
            dict | None: Parsed JSON body containing order statuses on success,
            or None if an error occurred while retrieving statuses.

        """
        try:
            await self.session.page.get(order_status(account_id))
            await self.session.page.sleep(2)

            url = order_info()

            async with self.session.page.expect_response(url) as response_info:
                await self.session.page.reload()

                # Wait for the response (with built-in timeout handling)
                await asyncio.wait_for(response_info.value, timeout=10)

                # Get response body directly
                body_str, _ = await response_info.response_body
            return json.loads(body_str)
        except Exception as e:
            print(f"Error getting order statuses: {e}")
            return None



import sys

from chase import account as acc
from chase import order as och
from chase import session
from chase import symbols as sym

# create Session Headless does not work at the moment it must be set to false.
cs = session.ChaseSession(
    title="Title of your profile here", headless=False, profile_path="your/profile/path"
)

# Login to Chase.com
login_one = cs.login("your_username", "your_password", "last_four_of_your_cell_phone")

# Check if login succeeded without needing 2fa if not then prompt for 2fa code
if login_one is False:
    print("Login succeeded without needing 2fa...")
else:
    code = input("Please input code that was sent to your phone: ")
    login_two = cs.login_two(code)

# Make all account object
all_accounts = acc.AllAccount(cs)

if all_accounts.account_connectors is None:
    sys.exit("Failed to get account connectors exiting script...")

# Get Account Identifiers
print("====================================")
print(f"Account Identifiers: {all_accounts.account_connectors}")

# Get Base Account Details
account_ids = list(all_accounts.account_connectors.keys())

print("====================================")
print("ACCOUNT DETAILS")
print("====================================")
for account in account_ids:
    account = acc.AccountDetails(account, all_accounts)
    print(account.nickname, account.mask, account.account_value)
print("====================================")

# Get Holdings
print("====================================")
print("HOLDINGS")
for account in account_ids:
    print("====================================")
    print(f"Account: {all_accounts.account_connectors[account]}")
    symbols = sym.SymbolHoldings(account, cs)
    success = symbols.get_holdings()
    if success:
        for i, symbol in enumerate(symbols.positions):
            if symbols.positions[i]["instrumentLongName"] == "Cash and Sweep Funds":
                symbol = symbols.positions[i]["instrumentLongName"]
                value = symbols.positions[i]["marketValue"]["baseValueAmount"]
                print(f"Symbol: {symbol} Value: {value}")
            elif symbols.positions[i]["assetCategoryName"] == "EQUITY":
                try:
                    symbol = symbols.positions[i]["positionComponents"][0][
                        "securityIdDetail"
                    ][0]["symbolSecurityIdentifier"]
                    value = symbols.positions[i]["marketValue"]["baseValueAmount"]
                    quantity = symbols.positions[i]["tradedUnitQuantity"]
                    print(f"Symbol: {symbol} Value: {value} Quantity: {quantity}")
                except KeyError:
                    symbol = symbols.positions[i]["securityIdDetail"]["cusipIdentifier"]
                    value = symbols.positions[i]["marketValue"]["baseValueAmount"]
                    quantity = symbols.positions[i]["tradedUnitQuantity"]
                    print(f"Symbol: {symbol} Value: {value} Quantity: {quantity}")
    else:
        print(f"Failed to get holdings for account {account}")
print("====================================")

# Create Order Object
order = och.Order(cs)

# Get Order Statuses
print("====================================")
print("ORDER STATUSES")
for account in account_ids:
    order_statuses = order.get_order_statuses(account)
    print("====================================")
    print(f"Account: {all_accounts.account_connectors[account]}")
    for order_status in order_statuses["orderSummaries"]:
        order_number = order_status["orderIdentifier"]
        order_type = order_status["tradeActionCode"]
        order_status_code = order_status["orderStatusCode"]
        print(
            f"Order Number: {order_number} Side: {order_type} Status: {order_status_code}"
        )
print("====================================")

# Get quote for INTC
symbol_quote = sym.SymbolQuote(account_ids[0], cs, "INTC")
print("====================================")
print("SYMBOL QUOTE")
print(
    f"{symbol_quote.security_description_text} ask price {symbol_quote.ask_price}, @{symbol_quote.as_of_timestamp} and the last trade was {symbol_quote.last_trade_price_amount}.")
print("====================================")

# Place dry run order for INTC
messages = order.place_order(
    account_ids[0],
    1,
    och.PriceType.MARKET,
    "INTC",
    och.Duration.DAY,
    och.OrderSide.BUY,
    dry_run=True,
)
if messages["ORDER CONFIRMATION"] != "":
    print(messages["ORDER CONFIRMATION"])
else:
    print(messages)



import asyncio
import json
import secrets
import traceback
from typing import TYPE_CHECKING

import anyio
import zendriver as uc

if TYPE_CHECKING:
    from zendriver import Browser, Tab

from .urls import landing_page, login_page, opt_out_verification_page


class ChaseSession:
    """A class to manage a session with Chase.

    This class provides methods to initialize a browser with the necessary options, log into Chase, and perform other actions on the Chase website.

    Attributes:
        title (str): Denotes the name of the profile and if populated will make the session persistent.
        headless (bool): Whether the browser should run in headless mode.
        profile_path (str): The path to the user profile directory for the browser.
        driver: The browser instance used to interact with the Chase website.

    Methods:
        get_browser(): Initializes and returns a browser with the necessary options.
        login(username, password, last_four): Logs into Chase with the provided credentials.
        login_two(code): Logs into Chase with the provided two-factor authentication code.
        save_storage_state(): Saves the storage state of the browser to a file.
        close_browser(): Closes the browser.

    """

    def __init__(self, *, docker: bool = False, headless: bool = True, title: str | None = None, profile_path: str = ".", debug: bool = False) -> None:
        """Initialize a new instance of the ChaseSession class.

        Args:
            docker (bool, optional): Whether the browser is running in a Docker environment. Defaults to False.
            headless (bool, optional): Whether the browser should run in headless mode. Defaults to True.
            title (string): Denotes the name of the profile and if populated will make the session persistent.
            profile_path (str, optional): The path to the user profile directory for the browser. Defaults to None.
            debug (bool, optional): Enable debug mode. Defaults to False.

        """
        self.docker: bool = docker
        self.headless: bool = headless
        self.title: str | None = title
        self.profile_path: str = profile_path
        self.password: str = ""
        self.browser: Browser
        self.page: Tab
        self.debug: bool = debug
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.loop.run_until_complete(self.get_browser())

    async def get_browser(self) -> None:
        """Initialize a browser instance using zendriver."""
        profile = await anyio.Path(self.profile_path).resolve() / self.title if self.title else await anyio.Path(self.profile_path).resolve() / "ZenChase"

        # keep self.profile_path as a string for backward compatibility with other code
        self.profile_path = str(profile)

        if not await profile.exists():
            await profile.parent.mkdir(parents=True, exist_ok=True)

        browser_args = []

        if self.docker:
            browser_args.extend(["--no-sandbox", "--disable-dev-shm-usage", "--disable-gpu", "--window-size=1920,1080"])
        elif self.headless:
            browser_args.extend(["--headless=new", "--window-size=1920,1080",
            "--user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
            "--disable-site-isolation-trials",
            "--disable-features=IsolateOrigins,site-per-process",
            "--disable-session-crashed-bubble",
            "--disable-infobars",
            "--disable-features=TranslateUI,VizDisplayCompositor",
            "--no-first-run",
            "--disable-default-apps",
            "--disable-extensions",
            "--no-sandbox",
            "--disable-dev-shm-usage",
            "--disable-gpu",
            "--window-size=1920,1080"])
        else:
            browser_args.extend([
                "--start-maximized",
                "--disable-session-crashed-bubble",
                "--user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
                "--disable-infobars",
                "--disable-features=TranslateUI,VizDisplayCompositor",
                "--no-first-run",
                "--disable-default-apps",
                "--disable-extensions",
            ])
        # zendriver has built-in stealth and anti-detection
        self.browser = await uc.start(
            browser_args=browser_args,
            user_data_dir=str(self.profile_path) if self.title else None,
        )
        self.page = await self.browser.get()

    def close_browser(self) -> None:
        """Close the browser.

        Raises:
            RuntimeError: If an unexpected runtime error occurs during browser closure.

        """
        try:
            return self.loop.run_until_complete(self._close_browser_async())
        except RuntimeError as e:
            if "This event loop is already running" in str(e):
                # We're in an async context, schedule the task
                return asyncio.create_task(self._close_browser_async())
            # Some other RuntimeError, re-raise it
            raise

    async def _close_browser_async(self) -> None:
        """Async implementation of close browser."""
        if self.browser:
            await self.browser.stop()

    def login(self, username: str, password: str, last_four: int) -> bool:
        """Log into the website with the provided username and password.

        Args:
            username (str): The user's username.
            password (str): The user's password.
            last_four (int): The last four digits of the user's phone number.

        Returns:
            bool: True if 2FA is required, False otherwise.

        """
        return self.loop.run_until_complete(
            self._login_async(username, password, last_four),
        )

    async def _login_async(self, username: str, password: str, last_four: int) -> bool:
        """Async implementation of login.

        Args:
            username (str): The user's username.
            password (str): The user's password.
            last_four (int): The last four digits of the user's phone number.

        Returns:
            bool: True if 2FA is required, False otherwise.

        Raises:
            Exception: If an unexpected error occurs during the login process.

        """
        try:
            self.password = r"" + password
            await self.page.get(login_page())
            await self.page.sleep(2)

            username_box = await self.page.find("#userId-input-field-input", timeout=30)
            password_box = await self.page.find("#password-input-field-input", timeout=30)

            if not username_box or not password_box:
                raise Exception("Could not find username or password fields.")

            # Scroll to help defeat bot detection?
            await self.page.scroll_down(int(secrets.SystemRandom().uniform(0, 40)), int(secrets.SystemRandom().uniform(900, 1500)))
            for letter in r"" + username:
                await username_box.send_keys(letter)
                await self.page.sleep(secrets.SystemRandom().uniform(0.05, 0.50))
            # Random mouse click to help defeat bot detection?
            await self.page.mouse_click(x=int(secrets.SystemRandom().uniform(0, 400)), y=int(secrets.SystemRandom().uniform(0, 400)), modifiers=0)
            # Scroll to help defeat bot detection?
            await self.page.scroll_up(int(secrets.SystemRandom().uniform(0, 40)), int(secrets.SystemRandom().uniform(900, 1500)))
            for letter in self.password:
                await password_box.send_keys(letter)
                await self.page.sleep(secrets.SystemRandom().uniform(0.05, 0.50))
            # Random mouse click to help defeat bot detection?
            await self.page.mouse_click(x=int(secrets.SystemRandom().uniform(0, 400)), y=int(secrets.SystemRandom().uniform(0, 400)), modifiers=0)
            signin_button = await self.page.find("#signin-button", timeout=5)
            await signin_button.click()
            await self.page.sleep(3)

            # Check if we are on the landing page (successful login without 2FA)
            if landing_page() in self.page.url:  # type: operator
                return False  # No 2FA needed

            # Check for 2FA options
            try:
                options_list = await self.page.find("#optionsList", timeout=15)
                if options_list:
                    # Handle shadow DOM
                    shadow_elements = await self.page.select_all(
                        "#optionsList *", timeout=5,
                    )
                    for element in shadow_elements:
                        attrs = element.attrs
                        text = attrs.get("label") if attrs else None
                        if text and ("Get a text" in text or "push notification" in text):
                            # Use JavaScript to click instead of element.click()
                            await element.apply("el => el.click()")
                            break
                    await self.page.sleep(1)

                    # Check for push notification
                    try:
                        approve_text = await self.page.find("text=approve", timeout=2)
                        if approve_text:
                            print(
                                "Chase is asking for 2FA from the phone app. You have 120 seconds to approve it.",
                            )
                            # Wait for redirect to landing page
                            for _ in range(120):
                                await self.page.sleep(1)
                                if landing_page() in self.page.url:
                                    return False
                    except TimeoutError:
                        # Timed out waiting for push notification moving on to text message flow
                        pass
                    try:
                        # Text message flow
                        radio_group = await self.page.find("mds-radio-group", timeout=15)
                        if radio_group:
                            attrs = radio_group.attrs
                            radio_buttons_json = attrs.get("radio-buttons") if attrs else None
                            if radio_buttons_json:
                                radio_buttons = json.loads(radio_buttons_json)
                                # Find the index of the button matching last_four
                                for index, button in enumerate(radio_buttons):
                                    if f"xxx-xxx-{last_four}" in button.get("label", ""):
                                        # Set the selected-index attribute using JavaScript
                                        await radio_group.apply(f"""
                                            el => {{
                                                el.setAttribute('selected-index', '{index}');
                                                el.dispatchEvent(new Event('change', {{ bubbles: true }}));
                                            }}
                                        """)
                                        await self.page.sleep(0.5)
                                        print(f"Selected radio button: {button['label']}")
                                        break
                            next_btn = await self.page.find("#next-content", timeout=5)
                            await next_btn.click()
                            return True  # 2FA code will be needed
                    except TimeoutError:
                        # Timed out waiting for text message option moving on to old 2fa flow.
                        pass

            except TimeoutError:
                # New 2FA options not found, proceed to old 2FA flow
                try:
                    dropdown = await self.page.find(
                        "#header-simplerAuth-dropdownoptions-styledselect", timeout=5,
                    )
                    await dropdown.click()
                    options = await self.page.query_selector_all('li[role="presentation"]')
                    for item in options:
                        text = await item.text
                        if str(last_four) in text:
                            await item.click()
                            break
                    submit_btn = await self.page.find('button[type="submit"]', timeout=5)
                    await submit_btn.click()
                    return True
                except asyncio.TimeoutError:
                    pass

            # Check for opt-out page
            if opt_out_verification_page() in self.page.url:
                # Trying to verify with xpath. This is untested in zendriver.
                skip_btn = await self.page.find("//button[contains(., 'Skip this step next time')]", timeout=5)
                await skip_btn.click()

                save_btn = await self.page.find("//button[contains(., 'Save and go to account')]", timeout=5)
                await save_btn.click()

            # Final check
            if landing_page() in self.page.url:
                return False

            raise Exception("Login failed due to an unknown page state.")

        except Exception as e:
            await self.close_browser()
            traceback.print_exc()
            raise Exception(f"Error in first step of login into Chase: {e}")

    def login_two(self, code: str) -> bool:
        """Perform the second step of login if 2FA needed.

        Args:
            code (str): 2FA code sent to users phone.

        Returns:
            bool: True if login is successful, False otherwise.

        """
        return self.loop.run_until_complete(self._login_two_async(code))

    async def _login_two_async(self, code: str) -> bool:
        """Async perform the second step of login if 2FA needed.

        Returns:
            bool: True if login is successful, False otherwise.

        Raises:
            TimeoutError: If landing page is not reached within 60 seconds.

        """
        try:
            code = str(code)
            await self.page.sleep(2)

            try:
                code_entry = await self.page.find("#otpInput", timeout=15)
                if code_entry:
                    await code_entry.send_keys(code)
                    next_btn = await self.page.find("#next-content", timeout=5)
                    await next_btn.click()
            except TimeoutError:
                pass

            try:
                # This is an old 2fa login flow untested in zendriver and probably obsolete
                otp_field = await self.page.find("#otpcode_input-input-field", timeout=15)
                await otp_field.send_keys(code)
                pwd_field = await self.page.find("#password_input-input-field", timeout=5)
                await pwd_field.send_keys(self.password)
                submit_btn = await self.page.find('button[type="submit"]', timeout=5)
                await submit_btn.click()
            except TimeoutError:
                pass

            try:
                await self.page.sleep(2)
                # Check for opt-out page
                if opt_out_verification_page() in self.page.url:
                    # Trying to verify with xpath. This is untested in zendriver.
                    skip_btn = await self.page.find("//button[contains(., 'Skip this step next time')]", timeout=5)
                    await skip_btn.click()

                    save_btn = await self.page.find("//button[contains(., 'Save and go to account')]", timeout=5)
                    await save_btn.click()
            except TimeoutError:
                pass

            # Wait for landing page
            for _ in range(60):
                await self.page.sleep(1)
                if landing_page() in self.page.url:
                    return True

            error_msg = "Failed to login to Chase. Landing page not reached after 60 seconds."
            raise TimeoutError(error_msg)

        except Exception as e:
            await self.close_browser()
            traceback.print_exc()
            print(f"Error logging into Chase: {e}")
            return False



import asyncio
import json
import traceback

from .session import ChaseSession
from .urls import account_info


class AllAccount:
    """A class to manage all accounts associated with a ChaseSession.

    This class provides methods to retrieve and manage information about all accounts associated with a given session.

    Attributes:
        session (ChaseSession): The session associated with the accounts.
        all_account_info (dict): Information about all accounts associated with the session.
        account_connectors (dict): Connectors for all accounts associated with the session.
        total_value (float): The total value of all accounts.
        total_value_change (float): The change in total value of all accounts.

    Methods:
        get_all_account_info(): Retrieves information about all accounts associated with the session.
        get_account_connectors(): Retrieves connectors for all accounts associated with the session.

    """

    def __init__(self, session: ChaseSession) -> None:
        """Initialize the AllAccount object with the given ChaseSession.

        This method initializes an AllAccount object, sets the session attribute to the given ChaseSession,
        and retrieves the account information and connectors for all accounts associated with the session.

        Args:
            session (ChaseSession): The session associated with the accounts.

        """
        self.session = session
        self.total_value = None
        self.total_value_change = None
        self.all_account_info = self.session.loop.run_until_complete(
            self.get_all_account_info(),
        )
        self.account_connectors = self.get_account_connectors()

    async def get_all_account_info(self) -> dict | None:
        """Retrieve and return information about all accounts associated with the session.

        This method navigates to the landing page, waits for the account information to load,
        and then retrieves the account information from the page.

        Returns:
            dict: A dictionary containing the account information, or None if the information could not be retrieved.

        """
        try:
            invest_json = await self.get_investment_json(account_info())
        except Exception as e:
            print(f"Timed out waiting for page to load: {e}")
            invest_json = None

        return invest_json

    def get_account_connectors(self) -> dict | None:
        """Retrieve and return connectors for all accounts associated with the session.

        This method iterates over all accounts in the all_account_info attribute and creates a dictionary
        where the keys are account IDs and the values are lists containing the corresponding account masks.

        Returns:
            dict: A dictionary containing the account connectors, or None if the all_account_info attribute is None.

        """
        account_dict = {}
        if self.all_account_info is None:
            return None
        if type(self.all_account_info) is list:
            for item in self.all_account_info:
                account_dict[item["accountId"]] = [item["mask"]]
        else:
            info = self.all_account_info["investmentAccountDetails"]
            for item in info:
                account_dict[item["accountId"]] = [item["mask"]]

        return account_dict

    async def get_investment_json(self, url: str) -> dict | None:
        """Fetch investment data from a given URL using zendriver.

        Returns:
            dict | None: A dictionary containing the parsed investment JSON on success,
            or None if the request timed out or an error occurred while parsing.

        """
        try:
            async with self.session.page.expect_response(url) as response_info:
                await self.session.page.reload()

                # Wait for the response (with built-in timeout handling)
                await asyncio.wait_for(response_info.value, timeout=10)

                # Get response body directly
                body_str, _ = await response_info.response_body
                data = json.loads(body_str)

                for info in data.get("cache", []):
                    if info.get("url") == "/svc/rr/accounts/secure/overview/investment/v1/list":
                        invest_json = info["response"]["investmentAccountOverviews"][0]
                        self.total_value = invest_json["totalValue"]
                        self.total_value_change = invest_json["totalValueChange"]
                        return invest_json

        except TimeoutError:
            print("Timed out waiting for network response.")
            return None
        except Exception as e:
            print(f"Error parsing response body: {e}")
            traceback.print_exc()
            return None

        return None


class AccountDetails:
    """A class to manage the details of a specific account associated with a ChaseSession.

    This class provides methods to retrieve and manage information about a specific account associated with a given session.

    Attributes:
        all_account_info (dict): Information about all accounts associated with the session.
        account_id (str): The ID of the account.
        nickname (str): The nickname of the account.
        mask (str): The mask of the account.
        detail_type (str): The detail type of the account.
        account_value (float): The value of the account.
        account_value_change (float): The change in value of the account.
        eda (str): The EDA of the account.
        ira (bool): Whether the account is an IRA.
        view_balance (bool): Whether the balance of the account can be viewed.
        prior_year_ira (bool): Whether the account is a prior year IRA.
        show_xfer (bool): Whether the account shows transfers.

    Methods:
        get_account_details(): Retrieves and sets the details of the account.

    """

    def __init__(self, account_id: str, all_account: AllAccount) -> None:
        """Initialize an AccountDetails object with a given account ID and AllAccount object.

        Args:
            account_id (str): The ID of the account.
            all_account (AllAccount): The AllAccount object associated with the session.

        """
        self.all_account_info = all_account.all_account_info
        self.account_id: str = account_id
        self.nickname: str = ""
        self.mask: str = ""
        self.detail_type: str = ""
        self.account_value: float = -1
        self.account_value_change: float = -1
        self.eda: str = ""
        self.ira: bool = False
        self.view_balance: bool = False
        self.prior_year_ira: bool = False
        self.show_xfer: bool = False
        self.get_account_details()

    def get_account_details(self) -> None:
        """Retrieve and set the details of the account.

        This method iterates over all accounts in the all_account_info attribute and sets the attributes of the AccountDetails object to the details of the account with the matching account ID.

        """
        info = self.all_account_info if type(self.all_account_info) is list else self.all_account_info["investmentAccountDetails"]
        for item in info:
            if item.get("accountId") is None:
                item = item[0]
            if item["accountId"] == self.account_id:
                self.mask = item.get("mask", "")
                self.nickname = item.get("nickname", "")
                self.detail_type = item.get("detailType", "")
                self.account_value = float(item.get("accountValue", -1))
                self.account_value_change = float(item.get("accountValueChange", -1))
                self.eda = bool(item.get("eda", False))
                self.ira = bool(item.get("ira", False))
                self.view_balance = bool(item.get("viewBalance", False))
                self.prior_year_ira = bool(item.get("priorYearIra", False))
                self.show_xfer = bool(item.get("showXfer", False))



"""Stores all the urls for the chase website used in this api."""

# API Endpoints
# API_ACCOUNT_LIST = "https://secure.chase.com/svc/rl/accounts/secure/v1/dashboard/module/list" # noqa: ERA001
# API_POSITIONS = "https://secure.chase.com/svc/wr/dwm/secure/gateway/investments/servicing/inquiry-maintenance/digital-investment-positions/v2/positions"  # noqa: ERA001


def login_page() -> str:
    """Return the Chase login page URL.

    Returns:
        str: The URL of the Chase login page.

    """
    return "https://secure05c.chase.com/web/auth/#/logon/logon/chaseOnline"


def auth_code_page() -> str:
    """Return the Chase authentication code page URL.

    Returns:
        str: The URL of the Chase authentication code page.

    """
    return "https://secure05c.chase.com/web/auth/#/logon/recognizeUser/provideAuthenticationCode"


def landing_page() -> str:
    """Return the Chase landing page URL.

    Returns:
        str: The URL of the Chase landing page.

    """
    return "https://secure.chase.com/web/auth/dashboard#/dashboard/overview"


def opt_out_verification_page() -> str:
    """Return the Chase opt-out verification page URL.

    Returns:
        str: The URL of the Chase opt-out verification page.

    """
    return "https://secure05c.chase.com/web/auth/#/logon/recognizeUser/esasiOptout"


def account_info() -> str:
    """Return the Chase account information URL.

    Returns:
        str: The URL of the Chase account information page.

    """
    return "https://secure.chase.com/svc/rl/accounts/secure/v1/dashboard/module/list"


def account_holdings(account_id: str) -> str:
    """Return the URL for the account holdings page for a specific account.

    This function takes an account ID as an argument and returns the URL for the
    account holdings page for that account on the Chase website.

    Args:
        account_id (str): The ID of the account for which to generate the URL.

    Returns:
        str: The URL for the account holdings page for the specified account.

    """
    return f"https://secure.chase.com/web/auth/dashboard#/dashboard/oi-portfolio/positions/render;ai={account_id}"


def holdings_json() -> str:
    """Return the URL for the holdings JSON endpoint.

    Returns:
        str: The URL to fetch holdings in JSON format.

    """
    return "https://secure.chase.com/svc/wr/dwm/secure/gateway/investments/servicing/inquiry-maintenance/digital-investment-positions/v2/positions"


def order_page() -> str:
    """Return the order page URL.

    Return the URL for the order page on the Chase website; this variant does not
    require an account id and returns the generic equity order entry URL.

    Returns:
        str: The URL for the order page.

    """
    return "https://secure.chase.com/web/auth/dashboard#/dashboard/oi-trade/equity/entry"


def quote_url() -> str:
    """Return the URL for the quote endpoint.

    Returns:
        str: The URL to fetch equity quotes.

    """
    return "https://secure.chase.com/svc/wr/dwm/secure/gateway/investments/servicing/inquiry-maintenance/digital-equity-quote/v1/quotes"


def execute_order(order_type: str) -> str:
    """Return the URL for the order execution endpoint.

    Args:
        order_type (str): The type of order, either "buy" or "sell".

    Returns:
        str: The URL to execute the specified type of order.

    Raises:
        ValueError: If order_type is not "buy" or "sell".

    """
    if order_type.lower() == "buy":
        return "https://secure.chase.com/svc/wr/dwm/secure/gateway/investments/servicing/investor-servicing/digital-equity-trades/v1/buy-orders"
    elif order_type.lower() == "sell":  # noqa: RET505
        return "https://secure.chase.com/svc/wr/dwm/secure/gateway/investments/servicing/investor-servicing/digital-equity-trades/v1/sell-orders"
    else:
        error_msg = f"Invalid order_type '{order_type}'. Must be either 'buy' or 'sell'."
        raise ValueError(error_msg)


def validate_order(order_type: str) -> str:
    """Return the URL for the order validation endpoint.

    Args:
        order_type (str): The type of order, either "buy" or "sell".

    Returns:
        str: The URL to validate the specified type of order.

    Raises:
        ValueError: If order_type is not "buy" or "sell".

    """
    if order_type.lower() == "buy":
        return "https://secure.chase.com/svc/wr/dwm/secure/gateway/investments/servicing/investor-servicing/digital-equity-trades/v1/buy-order-validations"
    elif order_type.lower() in {"sell", "sell_all"}:  # noqa: RET505
        return "https://secure.chase.com/svc/wr/dwm/secure/gateway/investments/servicing/investor-servicing/digital-equity-trades/v1/sell-order-validations"
    else:
        error_msg = f"Invalid order_type '{order_type}'. Must be either 'buy' or 'sell'."
        raise ValueError(error_msg)


def order_analytics() -> str:
    """Return the URL for the order analytics raw events endpoint.

    Returns:
        str: The full URL to the public raw analytics events endpoint used by Chase.

    """
    return "https://secure.chase.com/events/analytics/public/v1/events/raw/"


def order_status(account_id: str) -> str:
    """Return the URL for the order status page for a specific account.

    This function takes an account ID as an argument and returns the URL for the
    order status page for that account on the Chase website.

    Args:
        account_id (str): The ID of the account for which to generate the URL.

    Returns:
        str: The URL for the order status page for the specified account.

    """
    return f"https://secure.chase.com/web/auth/dashboard#/dashboard/trade/order/status;ai={account_id};orderStatus=ALL"


def order_info() -> str:
    """Return the Chase digital trade orders summaries endpoint URL.

    Returns:
        str: The URL for fetching digital trade order summaries.

    """
    return "https://secure.chase.com/svc/wr/dwm/secure/gateway/investments/servicing/inquiry-maintenance/digital-trade-orders/v1/summaries"


def get_headers() -> dict[str, str]:
    """Return the default HTTP headers used for requests to Chase endpoints.

    Returns:
        dict[str, str]: A mapping of header names to header values.

    """
    return {
        "accept": "application/json, text/plain, */*",
        "content-type": "application/json",
        "referer": "https://secure.chase.com/web/auth/dashboard",
        "x-jpmc-csrf-token": "NONE",
        "x-jpmc-channel": "id=C30",
        "origin": "https://secure.chase.com",
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
    }
